package org.radrails.rails.internal.ui.editors.rhtml;

import org.eclipse.jface.action.IAction;
import org.eclipse.jface.action.IToolBarManager;
import org.eclipse.jface.resource.JFaceResources;
import org.eclipse.jface.resource.LocalResourceManager;
import org.eclipse.jface.resource.ResourceManager;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.DocumentEvent;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.IDocumentListener;
import org.eclipse.jface.viewers.DoubleClickEvent;
import org.eclipse.jface.viewers.IDoubleClickListener;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.StyledText;
import org.eclipse.swt.dnd.DND;
import org.eclipse.swt.dnd.DragSourceEvent;
import org.eclipse.swt.dnd.DragSourceListener;
import org.eclipse.swt.dnd.DropTargetEvent;
import org.eclipse.swt.dnd.DropTargetListener;
import org.eclipse.swt.dnd.TextTransfer;
import org.eclipse.swt.dnd.Transfer;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IActionBars;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.views.contentoutline.ContentOutlinePage;
import org.radrails.rails.internal.ui.RailsUILog;
import org.radrails.rails.internal.ui.RailsUIPlugin;
import org.radrails.rails.internal.ui.editors.rhtml.model.RHTMLNode;

/**
 * Calculates and renders the outline view for an RHTML Editor
 * 
 * @author brad
 */
public class RHTMLContentOutlinePage extends ContentOutlinePage {
	private RHTMLEditor editor;
	private TreeViewer viewer;
	private IEditorInput input;
	private Object[] expanded;
	private StyledText textWidget;
	private RHTMLTreeDoubleClickListener treeDoubleClickListener;
	private RHTMLDocumentListener documentListener;
	private RHTMLMouseListener widgetMouseListener;
	private RHTMLKeyListener widgetKeyListener;

	public RHTMLContentOutlinePage(RHTMLEditor editor, IEditorInput input, StyledText textWidget) {
		this.editor = editor;
		this.input = input;
		this.textWidget = textWidget;
	}

	public void createControl(Composite parent) {
		super.createControl(parent);

		viewer = getTreeViewer();
		viewer.setContentProvider(new RHTMLContentProvider());
		viewer.setLabelProvider(new RHTMLLabelProvider());
		viewer.addSelectionChangedListener(this);

		// create listeners
		treeDoubleClickListener = new RHTMLTreeDoubleClickListener();
		documentListener = new RHTMLDocumentListener();
		widgetMouseListener = new RHTMLMouseListener();
		widgetKeyListener = new RHTMLKeyListener();

		// register listeners
		viewer.addDoubleClickListener(treeDoubleClickListener);
//		int operations = DND.DROP_COPY | DND.DROP_MOVE;
//		viewer.addDropSupport(operations, new Transfer[] { TextTransfer.getInstance() },
//				new RHTMLDropListener());
//		viewer.addDragSupport(operations, new Transfer[] { TextTransfer.getInstance() },
//				new RHTMLDragSourceListener());
		getDocument().addDocumentListener(documentListener);
		textWidget.addMouseListener(widgetMouseListener);
		textWidget.addKeyListener(widgetKeyListener);

		viewer.setInput(editor.getModel(true));
	}

	public void setActionBars(IActionBars actionBars) {
		super.setActionBars(actionBars);
		IToolBarManager tbManager = actionBars.getToolBarManager();
		IAction collapseAll = new RHTMLOutlineCollapseAction(this);
		collapseAll.setToolTipText("Collapse all");
		collapseAll.setImageDescriptor(RailsUIPlugin.getImageDescriptor("icons/collapseall.gif"));
		tbManager.add(collapseAll);
	}
	
	public void collapseOutline() {
		viewer.collapseAll();
	}

	private IDocument getDocument() {
		return editor.getDocumentProvider().getDocument(input);
	}

	public void dispose() {
		super.dispose();
		viewer.removeSelectionChangedListener(this);

		getDocument().removeDocumentListener(documentListener);
		textWidget.removeMouseListener(widgetMouseListener);
		textWidget.removeKeyListener(widgetKeyListener);
		viewer.removeDoubleClickListener(treeDoubleClickListener);
	}

	public void selectionChanged(SelectionChangedEvent event) {
		if (event.getSource() == getTreeViewer()) {
			StructuredSelection sel = (StructuredSelection) event.getSelection();
			RHTMLNode node = (RHTMLNode) sel.getFirstElement();
			if (node != null) {
				textWidget.setSelection(node.getStart());
//				textWidget.
				textWidget.setFocus();
			}
		}
	}

	/**
	 * Finds the node where the cursor currently is and highlights it in the
	 * outline page.
	 */
	private void highlightNodeAtCaret() {
		Display display = PlatformUI.getWorkbench().getDisplay();
		final ISelectionChangedListener outlinePage = this;
		display.asyncExec(new Runnable() {
			public void run() {
				int caret = textWidget.getCaretOffset();
				RHTMLNode selectedNode = getNodeAtCaretOffset(editor.getModel(false), caret);
				if (selectedNode != null) {
					StructuredSelection selection = new StructuredSelection(selectedNode);
					// TODO this seems like a bit of a hack :-( Can it fixed
					// with a bit of a reorg?
					viewer.removeSelectionChangedListener(outlinePage);
					viewer.setSelection(selection, true);
					viewer.addSelectionChangedListener(outlinePage);
				}
			}
		});
	}

	/**
	 * Find and return the node the caret is currently in.
	 * 
	 * @param nodes
	 *            the nodes to search through
	 * @param offset
	 *            the offset of the caret
	 * @return the node containing the caret, or null if it's not in any of the
	 *         given nodes
	 */
	public RHTMLNode getNodeAtCaretOffset(RHTMLNode[] nodes, int offset) {
		RHTMLNode result = null;

		for (int i = 0; result == null && i < nodes.length; i++) {
			int nodeStart = nodes[i].getStart();
			int nodeEnd = nodeStart + nodes[i].getLength();

			if (nodeStart > offset) {
				// then we've gone too far
				break;
			}
			else if (offset >= nodeStart && offset <= nodeEnd) {
				// then the node is somewhere in here
				if (nodes[i].hasChildren()) {
					result = getNodeAtCaretOffset((RHTMLNode[]) nodes[i].getChildren(), offset);
				}

				if (result == null) {
					result = nodes[i];
				}
			}
		}

		return result;
	}

	public class RHTMLLabelProvider extends LabelProvider {
		private ResourceManager fManager;
		
		public RHTMLLabelProvider() {
			fManager = new LocalResourceManager(JFaceResources.getResources());
		}
		
		public String getText(Object element) {
			if (element instanceof RHTMLNode) {
				return ((RHTMLNode) element).getLabel();
			}
			else {
				return super.getText(element);
			}
		}

		public Image getImage(Object element) {
			if (element instanceof RHTMLNode) {
				return fManager.createImageWithDefault(((RHTMLNode) element).getImage());
			}
			else {
				return super.getImage(element);
			}
		}
		public void dispose() {
			fManager.dispose();
			super.dispose();
		}
	}

	public class RHTMLTreeDoubleClickListener implements IDoubleClickListener {
		public void doubleClick(DoubleClickEvent event) {
			if (event.getSource() == getTreeViewer()) {
				StructuredSelection sel = (StructuredSelection) event.getSelection();
				RHTMLNode node = (RHTMLNode) sel.getFirstElement();
				if (node != null) {
					// select the node in the text editor
					editor.selectAndReveal(node.getStart(), node.getLength());
				}
			}
		}
	}

	public class RHTMLDocumentListener implements IDocumentListener {

		public void documentAboutToBeChanged(DocumentEvent event) {
			// keep track of what is visible in the tree
			expanded = viewer.getExpandedElements();
		}

		public void documentChanged(DocumentEvent event) {
			// refresh the tree and ensure what used to be visible is visible
			// now
			RHTMLNode[] oldModel = editor.getModel(false);
			RHTMLNode[] newModel = editor.getModel(true);
			if (newModel != null && !newModel.equals(oldModel)) {
				viewer.setInput(newModel);
				for (int i = 0; i < expanded.length; i++) {
					viewer.reveal(expanded[i]);
				}
			}
		}
	}

	public class RHTMLMouseListener implements MouseListener {
		public void mouseDoubleClick(MouseEvent e) {
		}

		public void mouseDown(MouseEvent e) {
		}

		/**
		 * Fired when a mouse button is released on the text widget
		 */
		public void mouseUp(MouseEvent e) {
			if (e.button == 1) {
				highlightNodeAtCaret();
			}
		}
	}

	public class RHTMLKeyListener implements KeyListener {
		private Thread timer;

		public void keyPressed(KeyEvent e) {
		}

		/**
		 * Fired when a key is released on the text widget
		 */
		public void keyReleased(KeyEvent e) {
			if (e.stateMask == 0
					&& (e.keyCode == SWT.ARROW_LEFT || e.keyCode == SWT.ARROW_RIGHT
							|| e.keyCode == SWT.ARROW_UP || e.keyCode == SWT.ARROW_DOWN)) {
				resetTimer();
			}
			else if (e.character == '>' && (e.stateMask == 0 || e.stateMask == SWT.SHIFT)) {
				highlightNodeAtCaret();
			}
		}

		private void resetTimer() {
			if (timer != null) {
				timer.interrupt();
			}
			
			timer = new Thread() {
				public void run() {
					try {
						sleep(1000);
						highlightNodeAtCaret();
					}
					catch (InterruptedException e) {
						// do nothing
					}
				}
			};
			
			timer.start();
		}
	}

	public class RHTMLDragSourceListener implements DragSourceListener {
		public void dragFinished(DragSourceEvent event) {
			StructuredSelection selection = (StructuredSelection) viewer.getSelection();
			if (selection.size() > 0) {
				RHTMLNode nodeToRemove = (RHTMLNode) selection.getFirstElement();
				try {
					getDocument().replace(nodeToRemove.getStart(), nodeToRemove.getLength(), "");
				}
				catch (BadLocationException e) {
					RailsUILog.logError(e.getMessage(), e);
				}
			}
		}

		public void dragSetData(DragSourceEvent event) {
			if (TextTransfer.getInstance().isSupportedType(event.dataType)) {
				StructuredSelection selection = (StructuredSelection) viewer.getSelection();
				if (selection.size() > 0) {
					event.data = ((RHTMLNode) selection.getFirstElement()).getCompleteText();
				}
				else {
					event.doit = false;
				}
			}
		}

		public void dragStart(DragSourceEvent event) {
		}
	}

	public class RHTMLDropListener implements DropTargetListener {

		public void dragEnter(DropTargetEvent event) {
		}

		public void dragLeave(DropTargetEvent event) {
		}

		public void dragOperationChanged(DropTargetEvent event) {
		}

		public void dragOver(DropTargetEvent event) {
		}

		public void drop(DropTargetEvent event) {
			if (event.item == null || event.item.getData() instanceof RHTMLNode) {
				IDocument document = getDocument();
				try {
					// insert provided data at the end of the document...
					int start = document.getLength();
					// unless an item is selected...
					if (event.item != null) {
						// in which case insert the provided data before the
						// currently selected node
						RHTMLNode targetNode = (RHTMLNode) event.item.getData();
						start = targetNode.getStart();
					}
					
					document.replace(start, 0, (String) event.data);
					textWidget.setCaretOffset(start);
				}
				catch (BadLocationException e) {
					RailsUILog.logError(e.getMessage(), e);
				}

			}
		}

		public void dropAccept(DropTargetEvent event) {
		}

	}
}
