package org.radrails.rails.internal.ui.editors.rhtml;

import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.eclipse.text.edits.ReplaceEdit;
import org.radrails.rails.internal.ui.RailsUIPlugin;
import org.radrails.rails.internal.ui.editors.rhtml.model.CommentNode;
import org.radrails.rails.internal.ui.editors.rhtml.model.DoctypeNode;
import org.radrails.rails.internal.ui.editors.rhtml.model.ERbNode;
import org.radrails.rails.internal.ui.editors.rhtml.model.RHTMLNode;
import org.radrails.rails.internal.ui.editors.rhtml.model.TagNode;
import org.radrails.rails.internal.ui.editors.rhtml.model.TextNode;
import org.rubypeople.rdt.core.formatter.CodeFormatter;
import org.rubypeople.rdt.internal.core.RubyModelManager;
import org.rubypeople.rdt.internal.corext.util.CodeFormatterUtil;

public class RHTMLFormatterVisitor extends RHTMLNodeVisitor {

	public static final String NEW_LINE = System.getProperty("line.separator");
	
	protected StringBuffer fBuffer;
	protected String fIndent;
	
	public RHTMLFormatterVisitor(String indent) {
		fBuffer = new StringBuffer();
		fIndent = indent;
	}
	
	public String getFormatted() {
		return fBuffer.toString();
	}
	
	protected boolean checkIndent(RHTMLNode parent, RHTMLNode child) {
		String parentKey = getPrefKey(parent);
		String childKey = getPrefKey(child);
		String prefKey = RHTMLFormatterConstants.getIndentPrefKey(parentKey, childKey);
		return RailsUIPlugin.getInstance().getPreferenceStore().getBoolean(prefKey);
	}
	
	private String getPrefKey(RHTMLNode node) {
		if(node instanceof TagNode) {
			return RHTMLFormatterConstants.NODE_TYPE_TAG;
		} else if(node instanceof ERbNode) {
			if(((ERbNode) node).isOutputNode()) {
				return RHTMLFormatterConstants.NODE_TYPE_ERB_OUTPUT;
			} else {
				return RHTMLFormatterConstants.NODE_TYPE_ERB_EVAL;
			}
		} else {
			return RHTMLFormatterConstants.NODE_TYPE_TEXT;
		}
	}
	
	protected String indent(String s, String currentIndent, boolean trim) {
		String[] lines = s.split(NEW_LINE);
		StringBuffer indented = new StringBuffer();

		for (int i = 0; i < lines.length; i++) {
			indented.append(currentIndent);
			String line = lines[i];
			if (trim) {
				line = line.trim();
			}
			indented.append(line);

			if (i < lines.length - 1) {
				indented.append(NEW_LINE);
			}
		}

		return indented.toString();
	}

	protected String incrementIndent(String currentIndent, String indentString) {
		currentIndent += indentString;

		return currentIndent;
	}
	
	protected String decrementIndent(String currentIndent, String indentString) {
		int indentLength = indentString.length();
		int currentIndentLength = currentIndent.length();
		if (currentIndentLength >= indentLength) {
			currentIndent = currentIndent.substring(0, currentIndentLength - indentLength);
		}

		return currentIndent;
	}
	
	// i feel like this method shouldn't be here...
	public void visit(RHTMLNode node, String currentIndent, boolean doIndent) {
		if(node instanceof ERbNode) {
			visit((ERbNode) node, currentIndent, doIndent);
		} else if(node instanceof TextNode) {
			visit((TextNode) node, currentIndent, doIndent);
		} else if(node instanceof TagNode) {
			visit((TagNode) node, currentIndent, doIndent);
		} else if(node instanceof CommentNode) {
			visit((CommentNode) node, currentIndent, doIndent);
		} else if(node instanceof DoctypeNode) {
			visit((DoctypeNode) node, currentIndent, doIndent);
		}
	}
	
	// convenience
	public void visit(RHTMLNode node, String currentIndent) {
		visit(node, currentIndent, true);
	}
	
	public void visit(CommentNode node, String currentIndent, boolean doIndent) {
		String trimmed = node.getText().trim();
		
		if(doIndent) {
			fBuffer.append(NEW_LINE);
			fBuffer.append(currentIndent);
			fBuffer.append(trimmed);
		} else {
			fBuffer.append(trimmed);
		}
	}
	
	public void visit(DoctypeNode node, String currentIndent, boolean doIndent) {
		String trimmed = node.getText().trim();
		
		if(doIndent) {
			fBuffer.append(NEW_LINE);
			fBuffer.append(currentIndent);
			fBuffer.append(trimmed);
		} else {
			fBuffer.append(trimmed);
		}
	}
	
	public void visit(ERbNode node, String currentIndent, boolean doIndent) {
		if(doIndent) {
			fBuffer.append(NEW_LINE);
		}
		String er = node.getText();

		Pattern erbPattern = Pattern.compile("(<%[-=]?)(.*?)(-?%>)", Pattern.DOTALL);
		Matcher matcher = erbPattern.matcher(er);
		if (matcher.matches()) {
			Map opts = RubyModelManager.getRubyModelManager().getOptions();
			int indent = 0;
			String text = matcher.group(2).trim();

			try {
				ReplaceEdit formattedRuby = (ReplaceEdit) CodeFormatterUtil.format2(
						CodeFormatter.K_RUBY_SCRIPT, text, indent, null, opts);
				text = formattedRuby.getText();
			}
			catch (IllegalArgumentException e) {
				/*
				 * this is bad, but not so bad. it's thrown when the ruby
				 * formatter tries to format an incomplete erb section (like <%
				 * end %> just ignore it and pad some spacing around text
				 */
			}

			// split multi lines of code to ensure each line
			// is indented a bit extra
			String[] lines = text.split(NEW_LINE);
			if(doIndent) {
				fBuffer.append(currentIndent);
			}
			if (lines.length > 1) {
				fBuffer.append(matcher.group(1));
				fBuffer.append(NEW_LINE);

				String childIndent = incrementIndent(currentIndent, fIndent);
				fBuffer.append(indent(text, childIndent, false));

				fBuffer.append(NEW_LINE);
				fBuffer.append(matcher.group(3));
			}
			else {
				fBuffer.append( matcher.group(1) + " " + text + " " + matcher.group(3));
			}

		}
		else {
			// doesn't seem to match the regex, so just print
			// it out as is
			if(doIndent) {
				fBuffer.append(indent(er, currentIndent, true));
			} else {
				fBuffer.append(er);
			}
		}
		String childIndent = incrementIndent(currentIndent, fIndent);
		RHTMLNode[] children = (RHTMLNode[]) node.getChildren();
		boolean childrenIndented = false;
		for (int i = 0; children != null && i < children.length; i++) {
			boolean doChildIndent = checkIndent(node, children[i]);
			childrenIndented = childrenIndented || doChildIndent;
			visit(children[i], childIndent, doChildIndent);
		}

		if ((node.isOpenNode() || node.isIntermediateNode()) && node.getNeedsClose()) {
			if(children.length > 0) {
				if(childrenIndented) {
					fBuffer.append(NEW_LINE);
					fBuffer.append(currentIndent);
				}
			} else if(doIndent) {
				fBuffer.append(NEW_LINE);
				fBuffer.append(currentIndent);
			}
			fBuffer.append(node.getCloseNode());
		}
	}

	public void visit(TagNode node, String currentIndent, boolean doIndent) {
		if(doIndent) {
			fBuffer.append(NEW_LINE);
			fBuffer.append(currentIndent);
		}
		fBuffer.append(node.getText().trim());

		String childIndent = incrementIndent(currentIndent, fIndent);
		RHTMLNode[] children = (RHTMLNode[]) node.getChildren();
		boolean childrenIndented = false;
		for (int i = 0; children != null && i < children.length; i++) {
			boolean doChildIndent = checkIndent(node, children[i]);
			childrenIndented = childrenIndented || doChildIndent;
			visit(children[i], childIndent, doChildIndent);
		}

		if (node.isOpenTag()) {
			if(children.length > 0) {
				if(childrenIndented) {
					fBuffer.append(NEW_LINE);
					fBuffer.append(currentIndent);
				}
			} else if(doIndent) {
				fBuffer.append(NEW_LINE);
				fBuffer.append(currentIndent);
			}
			fBuffer.append(node.getCloseNode());
		}
	}

	public void visit(TextNode node, String currentIndent, boolean doIndent) {
		/*
		 * Returns this text. All surrounding whitespace is trimmed EXCEPT if the
		 * tag has a number of leading newlines. Then one is kept to aid seperation
		 * of content.
		 */
		
		Pattern MULTIPLE_WHITESPACE_LINES = Pattern
			.compile("\\s*\\n\\s*\\n\\s*", Pattern.MULTILINE);
		String text = node.getText();
		String trimmed = text.trim();
		
		if (trimmed.length() > 0) {
			if(doIndent) {
				fBuffer.append(NEW_LINE);
				fBuffer.append(currentIndent);
				fBuffer.append(trimmed);
			} else {
				fBuffer.append(trimmed);
			}
		} else if (MULTIPLE_WHITESPACE_LINES.matcher(text).matches()) {
			fBuffer.append(NEW_LINE);
		}
	}

}
