package org.radrails.rails.internal.ui.editors.rhtml;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Stack;
import java.util.regex.Matcher;

import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.Document;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.rules.IToken;
import org.eclipse.jface.text.rules.Token;
import org.radrails.rails.internal.ui.editors.rhtml.model.CommentNode;
import org.radrails.rails.internal.ui.editors.rhtml.model.DoctypeNode;
import org.radrails.rails.internal.ui.editors.rhtml.model.ERbNode;
import org.radrails.rails.internal.ui.editors.rhtml.model.RHTMLNode;
import org.radrails.rails.internal.ui.editors.rhtml.model.TagNode;
import org.radrails.rails.internal.ui.editors.rhtml.model.TextNode;

public class RHTMLParser {
	private RHTMLPartitionScanner scanner;
	private IDocument document;
	private IToken token;
	private StringBuffer currentText;
	private Stack stack;

	// private boolean includeWhiteSpace;

	/**
	 * Parses the given html and returns an array of the RHTMLNodes at root
	 * level
	 * 
	 * @param rhtml
	 *            the rhtml to parse
	 * @param includeWhiteSpace
	 *            if false, TextNodes of only whitespace are ignored
	 * @return the RHTMLNodes at root level
	 * @throws BadLocationException
	 */
	public RHTMLNode[] parse(String rhtml) throws InvalidRHTMLException, BadLocationException {
		document = new Document(rhtml);
		scanner = RHTMLPartitionScanner.getInstance();
		scanner.setRange(document, 0, rhtml.length());
		currentText = new StringBuffer();

		token = scanner.nextToken();
		RHTMLNode[] res = getNodes(null);

		RHTMLNode trailingText = flushCurrentText();
		if (trailingText != null) {
			ArrayList l = new ArrayList(res.length + 1);
			l.addAll(Arrays.asList(res));
			l.add(trailingText);
			res = (RHTMLNode[]) l.toArray(res);
		}
		return res;
	}
	
	/**
	 * Parses the given rhtml and returns the closing node matching the last open node in the
	 * document. Note that the rhtml passed to this method is usually assumed to be incomplete.
	 * 
	 * @param rhtml
	 * @return a matching closing node, or null if none is found
	 * @throws InvalidRHTMLException 
	 * @throws BadLocationException 
	 */
	public String getCloseNode(String rhtml) throws BadLocationException, InvalidRHTMLException {
		document = new Document(rhtml);
		scanner = RHTMLPartitionScanner.getInstance();
		scanner.setRange(document, 0, rhtml.length());
		currentText = new StringBuffer();
		stack = new Stack();

		token = scanner.nextToken();
		findOpenNodes(null);
		RHTMLNode topNode = null;
		if(!stack.isEmpty()) {
			topNode = (RHTMLNode) stack.pop();
		}
		return topNode == null ? null : topNode.getCloseNode();
	}
	
	private void findOpenNodes(RHTMLNode parent) throws BadLocationException, InvalidRHTMLException {
		while (token != Token.EOF) {
			RHTMLNode node = processToken();
			if (!(node instanceof TextNode)) {
				/*
				 * if it's a text node then the scanner has already been
				 * incremented. If not, we need to increment it manually so we
				 * will now.
				 */
				if(node instanceof ERbNode) {
					ERbNode erbNode = (ERbNode) node;
					if(!erbNode.isIntermediateNode() || !erbNode.isMatchingIntermediateNode(parent)) {
						token = scanner.nextToken();
					}
				} else {
					token = scanner.nextToken();
				}
			}

			if (node instanceof TagNode) {
				TagNode tagNode = (TagNode) node;
				if (tagNode.isCloseTag() && tagNode.isMatchingCloseTag(parent)) {
					stack.pop();
					// tag is closed. we're done here
					break;
				}
				else if (tagNode.isOpenTag()) {
					stack.push(tagNode);
					findOpenNodes(tagNode);
//					tagNode.setChildren(children);
				}
			}
			
			if (node instanceof ERbNode) {
				ERbNode erbNode = (ERbNode) node;
				if(erbNode.isCloseNode() && erbNode.isMatchingCloseNode(parent)) {
					stack.pop();
					break;
				} else if (erbNode.isIntermediateNode() && erbNode.isMatchingIntermediateNode(parent)) {
					stack.pop();
					break;
				} else if (erbNode.isOpenNode() || erbNode.isIntermediateNode()) {
					stack.push(erbNode);
					findOpenNodes(erbNode);
//					erbNode.setChildren(children);
				}
			}

//			if (node != null) {
//				node.setParent(parent);
//				res.add(node);
//			}
		}
	}

	/**
	 * Gets the nodes until the close of the parent tag
	 * 
	 * @param parent
	 *            parent tag of the nodes
	 * @return
	 * @throws BadLocationException
	 */
	private RHTMLNode[] getNodes(RHTMLNode parent) throws InvalidRHTMLException,
			BadLocationException {
		ArrayList res = new ArrayList();

		while (token != Token.EOF) {
			RHTMLNode node = processToken();
			if (!(node instanceof TextNode)) {
				/*
				 * if it's a text node then the scanner has already been
				 * incremented. If not, we need to increment it manually so we
				 * will now.
				 */
				if(node instanceof ERbNode) {
					ERbNode erbNode = (ERbNode) node;
					if(!erbNode.isIntermediateNode() || !erbNode.isMatchingIntermediateNode(parent)) {
						token = scanner.nextToken();
					}
				} else {
					token = scanner.nextToken();
				}
			}

			if (node instanceof TagNode) {
				TagNode tagNode = (TagNode) node;
				if (tagNode.isCloseTag() && tagNode.isMatchingCloseTag(parent)) {
					// tag is closed. we're done here
					break;
				}
				else if (tagNode.isOpenTag()) {
					RHTMLNode[] children = getNodes(tagNode);
					tagNode.setChildren(children);
				}
			}
			
			if (node instanceof ERbNode) {
				ERbNode erbNode = (ERbNode) node;
				if(erbNode.isCloseNode() && erbNode.isMatchingCloseNode(parent)) {
					break;
				} else if (erbNode.isIntermediateNode() && erbNode.isMatchingIntermediateNode(parent)) {
					((ERbNode) parent).setNeedsClose(false);
					break;
				} else if (erbNode.isOpenNode() || erbNode.isIntermediateNode()) {
					RHTMLNode[] children = getNodes(erbNode);
					erbNode.setChildren(children);
				}
			}

			if (node != null) {
				node.setParent(parent);
				res.add(node);
			}
		}

		return (RHTMLNode[]) res.toArray(new RHTMLNode[0]);
	}

	/**
	 * Processes the latest token from the scanner
	 * 
	 * @return a RHTMLNode representing the token, or null if token is
	 *         incomplete
	 * @throws BadLocationException
	 * @throws InvalidRHTMLException
	 */
	private RHTMLNode processToken() throws BadLocationException, InvalidRHTMLException {
		Object data = token.getData();
		String text = document.get(scanner.getTokenOffset(), scanner.getTokenLength());
		RHTMLNode res = null;

		if (data != null) {
			RHTMLNode textNode = flushCurrentText();
			if (textNode != null) {
				res = textNode;
			}
			else if (data.equals(RHTMLPartitionScanner.RHTML_RUBY_EXPRESSION)) {
				res = processERbNode(text);
			}
			else if (data.equals(RHTMLPartitionScanner.RHTML_TAG)) {
				res = new TagNode(text, scanner.getTokenOffset());
			}
			else if (data.equals(RHTMLPartitionScanner.RHTML_MULTILINE_COMMENT)) {
				res = new CommentNode(text, scanner.getTokenOffset());
			}
			else if (data.equals(RHTMLPartitionScanner.RHTML_DOCTYPE)) {
				res = new DoctypeNode(text, scanner.getTokenOffset());
			}
		}
		else {
			currentText.append(text);
		}

		return res;
	}

	private RHTMLNode processERbNode(String text) throws BadLocationException {
		RHTMLNode res = null;
		// check if it's a one line comment and set as a comment node if
		// so
		int lineNumber = document.getLineOfOffset(scanner.getTokenOffset());
		int start = document.getLineOffset(lineNumber);
		String line = document.get(start, document.getLineLength(lineNumber)).trim();
		Matcher matcher = RHTMLToggleCommentAction.COMMENTED_LINE_PATTERN.matcher(line);
		if (matcher != null && matcher.matches()) {
			int lastTokenStart = scanner.getTokenOffset() + line.length();
			while (scanner.getTokenOffset() < lastTokenStart && scanner.nextToken() != Token.EOF) {
				text += document.get(scanner.getTokenOffset(), scanner.getTokenLength());
			}
			res = new CommentNode(text, start);
		}
		else {
			// otherwise it's just a normal ERbNode
			res = new ERbNode(text, scanner.getTokenOffset());
		}
		
		return res;
	}

	/**
	 * Text gets tokenised one char at a time, so we need to append to the
	 * currentText stringbuffer. Once a different token is encoutered (or rhtml
	 * ends), this method should o be called to flush the text.
	 */
	private RHTMLNode flushCurrentText() {
		RHTMLNode res = null;
		if (currentText.length() != 0) {
			String text = currentText.toString();
			res = new TextNode(text, scanner.getTokenOffset() - text.length());

			currentText = new StringBuffer();
		}

		return res;
	}

	public static class InvalidRHTMLException extends Exception {
		private static final long serialVersionUID = 1L;
		private String badNode;

		public InvalidRHTMLException(String badNode) {
			this.badNode = badNode;
		}

		public String getMessage() {
			return "Error parsing string with text - " + badNode;
		}
	}
}
