/*******************************************************************************
 * Copyright (c) 2005 RadRails.org and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.radrails.org/legal/cpl-v10.html
 *******************************************************************************/

package org.radrails.rails.internal.ui.editors.rhtml;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.BadPositionCategoryException;
import org.eclipse.jface.text.DocumentCommand;
import org.eclipse.jface.text.DocumentEvent;
import org.eclipse.jface.text.IAutoEditStrategy;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.IPositionUpdater;
import org.eclipse.jface.text.link.ILinkedModeListener;
import org.eclipse.jface.text.link.InclusivePositionUpdater;
import org.eclipse.jface.text.link.LinkedModeModel;
import org.eclipse.jface.text.link.LinkedModeUI;
import org.eclipse.jface.text.link.LinkedPosition;
import org.eclipse.jface.text.link.LinkedPositionGroup;
import org.eclipse.jface.text.source.IOverviewRuler;
import org.eclipse.jface.text.source.IVerticalRuler;
import org.eclipse.jface.text.source.SourceViewerConfiguration;
import org.eclipse.jface.text.source.projection.ProjectionViewer;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.swt.widgets.Composite;
import org.radrails.rails.internal.ui.RailsUIPlugin;
import org.radrails.rails.internal.ui.editors.TabExpander;
import org.radrails.rails.internal.ui.preferences.IRailsPreferenceConstants;

public class RHTMLSourceViewer extends ProjectionViewer {

	private boolean isTabReplacing = false;

	private boolean fIgnoreTextConverters = false;

	private TabExpander tabExpander;

	/** The auto edit category */
	protected static final String CATEGORY_AUTO_EDIT = "org.radrails.rails.ui"
			+ ".auto.edit." + System.currentTimeMillis(); //$NON-NLS-1$

	/** The position updater for auto edit */
	protected final IPositionUpdater fAutoEditUpdater;

	/** The linked position list for code auto edit */
	protected final LinkedList fPositionList = new LinkedList();

	private IPropertyChangeListener propListener = new IPropertyChangeListener() {
		public void propertyChange(PropertyChangeEvent event) {
			if (event.getProperty().equals(
					IRailsPreferenceConstants.RHTML_FORMAT_USE_TAB)) {
				isTabReplacing = !RailsUIPlugin.getInstance()
						.getPreferenceStore().getBoolean(
								IRailsPreferenceConstants.RHTML_FORMAT_USE_TAB);
				if (isTabReplacing) {
					int length = RailsUIPlugin
							.getInstance()
							.getPreferenceStore()
							.getInt(IRailsPreferenceConstants.RHTML_INDENTATION);
					updateIndentPrefixes(length);
				} else {
					updateIndentPrefixes();
				}
			} else if (event.getProperty().equals(
					IRailsPreferenceConstants.RHTML_INDENTATION)) {
				int length = RailsUIPlugin.getInstance().getPreferenceStore()
						.getInt(IRailsPreferenceConstants.RHTML_INDENTATION);
				tabExpander = new TabExpander(length);
				updateIndentPrefixes(length);
			}
		}
	};

	public RHTMLSourceViewer(Composite composite, IVerticalRuler verticalRuler,
			IOverviewRuler overviewRuler, boolean overviewRulerVisible,
			int styles) {
		super(composite, verticalRuler, overviewRuler, overviewRulerVisible,
				styles);
		initializeTabReplace();
		RailsUIPlugin.getInstance().getPreferenceStore()
				.addPropertyChangeListener(propListener);
		fAutoEditUpdater = new InclusivePositionUpdater(CATEGORY_AUTO_EDIT);
	}

	public void configure(SourceViewerConfiguration configuration) {
		super.configure(configuration);
		if (isTabReplacing) {
			int length = RailsUIPlugin.getInstance().getPreferenceStore()
					.getInt(IRailsPreferenceConstants.RHTML_INDENTATION);
			updateIndentPrefixes(length);
		} else {
			updateIndentPrefixes();
		}
	}

	public void doOperation(int operation) {
		if (getTextWidget() == null || !redraws()) {
			return;
		}

		switch (operation) {
		case UNDO:
			fIgnoreTextConverters = true;
			break;
		case REDO:
			fIgnoreTextConverters = true;
			break;
		}

		super.doOperation(operation);
	}

	protected void customizeDocumentCommand(DocumentCommand command) {
		if (isIgnoringAutoEditStrategies())
			return;

		List strategies = (List) selectContentTypePlugin(command.offset,
				fAutoIndentStrategies);
		if (strategies == null)
			return;

		IDocument document = getDocument();
		if (!strategies.isEmpty()) {

			fPositionList.clear();
			String originalCommandText = command.text;
			
			LinkedPosition[] result = null;
			IAutoEditStrategy strategy = null;

			for (final Iterator iterator = new ArrayList(strategies).iterator(); iterator
					.hasNext();) {

				strategy = (IAutoEditStrategy) iterator.next();
				strategy.customizeDocumentCommand(document, command);

				if ((strategy instanceof ILinkedModeEditStrategy) && originalCommandText.equals("\t")) {
					result = ((ILinkedModeEditStrategy) strategy)
							.getLinkedPositions();
					if (result != null && result.length > 0)
						fPositionList.addAll(Arrays.asList(result));
				}
			}
		}
		if (!fIgnoreTextConverters) {
			convertTabs(command, getDocument());
		}
		fIgnoreTextConverters = false;
	}

	void initializeTabReplace() {
		this.isTabReplacing = !RailsUIPlugin.getInstance().getPreferenceStore()
				.getBoolean(IRailsPreferenceConstants.RHTML_FORMAT_USE_TAB);
		if (this.isTabReplacing) {
			int length = RailsUIPlugin.getInstance().getPreferenceStore()
					.getInt(IRailsPreferenceConstants.RHTML_INDENTATION);
			tabExpander = new TabExpander(length);
		}
	}

	protected void convertTabs(DocumentCommand command, IDocument document) {
		if (!isTabReplacing)
			return;

		if (command.text == null)
			return;

		if (command.text.equals("\t"))
			tabExpander.expandTab(command, document);
	}

	public boolean isTabReplacing() {
		return isTabReplacing;
	}

	public String getIndentString() {
		return tabExpander.getFullIndent();
	}

	private void updateIndentPrefixes() {
		String[] prefixes = new String[] { "\t", " ", "" };
		setIndentPrefixesForContentTypes(prefixes);
	}

	private void updateIndentPrefixes(int spacesPerTab) {
		StringBuffer spaces = new StringBuffer();
		for (int i = 0; i < spacesPerTab; i++) {
			spaces.append(" ");
		}
		String[] prefixes = new String[] { spaces.toString(), "\t", "" };
		setIndentPrefixesForContentTypes(prefixes);
	}

	private void setIndentPrefixesForContentTypes(String[] prefixes) {
		setIndentPrefixes(prefixes, IDocument.DEFAULT_CONTENT_TYPE);
		String[] rhtmlTypes = RHTMLPartitionScanner.RHTML_PARTITION_TYPES;
		for (int i = 0; i < rhtmlTypes.length; i++) {
			setIndentPrefixes(prefixes, rhtmlTypes[i]);
		}
	}

	/*
	 * @see org.eclipse.jface.text.TextViewer#handleVisibleDocumentChanged(org.eclipse.jface.text.DocumentEvent)
	 */
	protected final void handleVisibleDocumentChanged(final DocumentEvent event) {
		super.handleVisibleDocumentChanged(event);

		if (!fPositionList.isEmpty()) {

			try {

				final IDocument document = event.getDocument();
				final LinkedModeModel model = new LinkedModeModel();

				final String category = CATEGORY_AUTO_EDIT;
				if (!document.containsPositionCategory(category)) {

					document.addPositionCategory(category);
					document.addPositionUpdater(fAutoEditUpdater);

					model.addLinkingListener(new ILinkedModeListener() {

						public final void left(final LinkedModeModel dummy,
								final int flags) {

							if (document.containsPositionCategory(category)) {

								try {
									document.removePositionCategory(category);
								} catch (BadPositionCategoryException exception) {
									// Do nothing
								}
								document
										.removePositionUpdater(fAutoEditUpdater);
							}
						}

						public final void resume(final LinkedModeModel dummy,
								final int flags) {
						}

						public final void suspend(final LinkedModeModel dummy) {
						}
					});
				}

				LinkedPosition position = null;
				LinkedPositionGroup group = null;

				for (final Iterator iterator = fPositionList.iterator(); iterator
						.hasNext();) {

					position = (LinkedPosition) iterator.next();

					group = new LinkedPositionGroup();
					group.addPosition(position);

					model.addGroup(group);
				}
				model.forceInstall();

				final LinkedModeUI handler = new LinkedModeUI(model, this);

				final LinkedPosition exit = (LinkedPosition) fPositionList
						.getFirst();
				final LinkedPosition entry = (LinkedPosition) fPositionList
						.getLast();

				addSelectionChangedListener(new ISelectionChangedListener() {

					private boolean fHandled = false;

					public final void selectionChanged(
							final SelectionChangedEvent dummy) {

						if (!fHandled) {

							fHandled = true;

							removeSelectionChangedListener(this);
						}
					}
				});

				addPostSelectionChangedListener(new ISelectionChangedListener() {

					private boolean fHandled = false;

					public final void selectionChanged(
							final SelectionChangedEvent dummy) {

						if (!fHandled) {
							setSelectedRange(entry.offset, entry.length);

							invalidateTextPresentation(entry.offset, 0);

							fHandled = true;

							removePostSelectionChangedListener(this);
						}
					}
				});

				handler
						.setExitPosition(this, exit.offset, 0,
								LinkedPositionGroup.NO_STOP);

				handler.enter();

			} catch (BadLocationException exception) {
				// Do nothing
			} finally {
				fPositionList.clear();
			}
		}
	}

}
