package org.radrails.rails.internal.ui.editors.rhtml;

import java.util.ResourceBundle;
import java.util.regex.Pattern;

import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.TextSelection;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.ui.texteditor.ITextEditor;
import org.eclipse.ui.texteditor.TextEditorAction;
import org.radrails.rails.internal.ui.RailsUILog;

/**
 * Toggles a comment in the RHTML editor
 * 
 * @author brad
 */
public class RHTMLToggleCommentAction extends TextEditorAction {
	private static final String NEW_LINE = System.getProperty("line.separator");
	public static final String COMMENT_START = "<% if false %>";
	public static final String COMMENT_END = "<% end %>";
	public static final Pattern COMMENTED_LINE_PATTERN = Pattern.compile("(\\w*)(" + COMMENT_START
			+ ")(.*)(" + COMMENT_END + ")(\\w*)");

	public RHTMLToggleCommentAction(ResourceBundle bundle, String prefix, ITextEditor editor) {
		super(bundle, prefix, editor);
	}

	public boolean isEnabled() {
		return true;
	}

	public void run() {
		IDocument doc = this.getTextEditor().getDocumentProvider().getDocument(
				this.getTextEditor().getEditorInput());
		ISelection selection = this.getTextEditor().getSelectionProvider().getSelection();
		try {
			if (selection instanceof TextSelection) {
				TextSelection textSelection = (TextSelection) selection;
				String text = textSelection.getText();
				int startPos = 0;
				int endPos = 0;

				if (text == null || text.length() == 0) {
					// just get the current line
					int startLine = textSelection.getStartLine();
					startPos = doc.getLineOffset(startLine);
					endPos = startPos + doc.getLineLength(startLine);
				}
				else {
					// get the whole of the selected lines
					startPos = doc.getLineOffset(textSelection.getStartLine());
					int endLine = textSelection.getEndLine();
					endPos = doc.getLineOffset(endLine) + doc.getLineLength(endLine);

				}
				text = doc.get(startPos, endPos - startPos);
				String toggled = toggleComment(text);
				doc.replace(startPos, endPos - startPos, toggled);
				this.getTextEditor().selectAndReveal(startPos, toggled.length());
			}

		}
		catch (BadLocationException e) {
			RailsUILog.logError(e.getMessage(), e);
		}
		super.run();
	}

	/**
	 * Comment the provide string so that no ERB will be run in it.
	 * 
	 * I used the standard eclipse java toggle comment as a guide:
	 * <ul>
	 * <li>if no line is selected, the current line is toggled</li>
	 * <li>if multiple lines are selected and all are commented, then they are
	 * all set to uncommented</li>
	 * <li>if multiple lines are selected and at least one is uncommented, then
	 * all lines are set to be commented (if a line is already commented, it is
	 * left as-is).</li>
	 * </ul>
	 * 
	 * @param s
	 *            The string to be toggled
	 * @return
	 */
	public static String toggleComment(String s) {
		String[] lines = s.split(NEW_LINE);
		StringBuffer toggled = new StringBuffer();
		boolean allCommented = true;

		for (int i = 0; allCommented && i < lines.length; i++) {
			allCommented &= COMMENTED_LINE_PATTERN.matcher(lines[i]).find();
		}

		for (int i = 0; i < lines.length; i++) {
			if (allCommented) {
				// remove the comments
				toggled.append(COMMENTED_LINE_PATTERN.matcher(lines[i]).replaceAll("$1$3$5"));
			}
			else if (!COMMENTED_LINE_PATTERN.matcher(lines[i]).find()) {
				// add a comment if it's missing
				toggled.append("<% if false %>");
				toggled.append(lines[i]);
				toggled.append("<% end %>");
			}
			else {
				// already commented so leave it alone
				toggled.append(lines[i]);
			}
			toggled.append(NEW_LINE);
		}

		return toggled.toString();
	}
}
