package org.radrails.rails.internal.ui.editors.rhtml.model;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.eclipse.jface.resource.ImageDescriptor;
import org.radrails.rails.internal.ui.RailsUIPlugin;
import org.radrails.rails.internal.ui.editors.rhtml.RHTMLPartitionScanner;
import org.radrails.rails.internal.ui.editors.rhtml.RHTMLParser.InvalidRHTMLException;

/**
 * Represents an HTML tag. Both normal tags with close tags (a, div, body) are
 * represented by this tag, as well as brs that have no matching close
 * 
 * @author brad
 * 
 */
public class TagNode extends RHTMLNode {
	private static final ImageDescriptor OUTLINE_IMAGE = RailsUIPlugin.getImageDescriptor(
			"icons/html_tag_obj.gif");

	private String tagType;

	public TagNode(String text, int start) throws InvalidRHTMLException {
		super(text, start, RHTMLPartitionScanner.RHTML_TAG);
		Matcher matcher = Pattern.compile("</?([^\\s]*)\\s?.*>").matcher(text);
		if (matcher.matches()) {
			tagType = matcher.group(1);
		}
		else {
			throw new InvalidRHTMLException(text);
		}
	}

	public String getCloseNode() {
		return "</" + getTagType() + ">";
	}

	public String getLabel() {
		return getTagType();
	}

	public ImageDescriptor getImage() {
		return OUTLINE_IMAGE;
	}

	public boolean isOpenTag() {
		String tag = getText().toLowerCase();
		return !isCloseTag() && !tag.endsWith("/>") && !tag.toLowerCase().equalsIgnoreCase("<br>");
	}

	public boolean isCloseTag() {
		String tag = getText().toLowerCase();
		return tag.startsWith("</");
	}

	public boolean isMatchingCloseTag(RHTMLNode parent) {
		boolean res = false;
		if (parent instanceof TagNode) {
			TagNode tnParent = (TagNode) parent;
			res = tnParent.isOpenTag() && tnParent.getTagType() != null && getTagType() != null
					&& tnParent.getTagType().equalsIgnoreCase(getTagType());
		}

		return res;
	}

	private String getTagType() {
		return tagType;
	}

	/**
	 * Gets the entire length of this tag - the length from the start of the
	 * start tag to the end of the close tag, including all child tags in
	 * between
	 */
	public int getLength() {
		int length = super.getLength();
		if (hasChildren()) {
			RHTMLNode[] children = (RHTMLNode[]) getChildren();
			RHTMLNode lastChild = children[children.length - 1];
			int end = lastChild.getStart() + lastChild.getLength() + getCloseNode().length();
			length = end - getStart();
		}

		return length;
	}

	/**
	 * Gets the complete text of this tag - from the start of the start to the
	 * end of the close tag, including all child tags in between
	 */
	public String getCompleteText() {
		StringBuffer res = new StringBuffer();
		res.append(getText());

		RHTMLNode[] children = (RHTMLNode[]) getChildren();
		for (int i = 0; children != null && i < children.length; i++) {
			res.append(children[i].getCompleteText());
		}

		if (isOpenTag()) {
			res.append(getCloseNode());
		}

		return res.toString();
	}
}
